package edu.uiuc.edsg;

import java.util.LinkedList;

/**
 * Traverses an iteration tree, moving all Wrappers to the end (left) of the iteration.
 */
public class WrapperReorderVisitor implements IIteratorVisitor {
	private IIterationLink _workingLink;
	private LinkedList<GeneratorLink> _wrapperList;
	
	private static final IIterationLink WRAPPER_SENTINEL = null; // null is the easiest sentinel 
	
	public WrapperReorderVisitor() {}
	
	public void visit(BinaryLink link) {
		link.getLeft().accept(this);
		IIterationLink left = _workingLink;
		link.getRight().accept(this);
		IIterationLink right = _workingLink;

		if (left == WRAPPER_SENTINEL && right == WRAPPER_SENTINEL) {
			// If both children were wrappers, then ignore them
			_workingLink = WRAPPER_SENTINEL;			
		}
		else if (left == WRAPPER_SENTINEL) {
			// If left was a wrapper, use right
			_workingLink = right;
		}
		else if (right == WRAPPER_SENTINEL) {
			// if right was a wrapper, use left
			_workingLink = left;
		}
		else {						
			// if neither child was a wrapper, create a new binary link
			//  with the new child trees
			_workingLink = link.createNewInstance(left, right);
		}
	}

	public void visit(CombinatoricLink link) {
		visit((BinaryLink)link);
	}

	public void visit(SequentialLink link) {
		visit((BinaryLink)link);
	}

	public void visit(GeneratorLink link) {
		if (link.getInnerGenerator() instanceof Wrapper) {
			_wrapperList.addFirst(link);
			// signal the parent that this generator was a wrapper
			_workingLink = WRAPPER_SENTINEL;
		}
		else if (link.getInnerGenerator() instanceof CompositeGenerator) {
			// expand tree to include composite generator's tree
			((CompositeGenerator)link.getInnerGenerator()).getIterator().accept(this);
		}
		else {
			_workingLink = link;
		}
	}
	
	/**
	 * @return The IIterationLink that iterates identically to the original 
	 *  iterator, but with all the wrappers moved to the end of the iteration.
	 */
	public IIterationLink getReorderedIterator(IIterationLink root) {
		_wrapperList = new LinkedList<GeneratorLink>();
		_workingLink = null;
		root.accept(this);
		for (GeneratorLink wrapper : _wrapperList) {
			// we use combinatoric link because it will only iterate the
			//  wrapper once as opposed to a sequential link which will
			//  reset and iterate the wrapper on every tree iteration 
			_workingLink = new CombinatoricLink(wrapper, _workingLink);
		}
		return _workingLink;
	}
}
